/*
 * Copyright 2023 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _CHROMOBIUS_COMMANDS_MAIN_ALL_TEST_H
#define _CHROMOBIUS_COMMANDS_MAIN_ALL_TEST_H

#include <stdexcept>
#include <string>
#include <vector>

#include "chromobius/commands/main_all.h"

namespace chromobius {

std::string result_of_running_main(const std::vector<std::string> args, const std::string &input);

}  // namespace chromobius

#endif
